/*
    Disk latency benchmarking tool
    Copyright (C) 2006-2008 Darrick Wong

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "bogoseek.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <inttypes.h>
#include <time.h>
#include <unistd.h>
#include <stdint.h>
#include <sched.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/syscall.h>
#include <asm/unistd.h>

static FILE *tty;
typedef ssize_t (*io_func_t)(int fd, void *buf, size_t count);
static io_func_t io_func = read;
static int access_flag = O_RDONLY;

#ifdef O_DIRECT
static int direct_flag = O_DIRECT;
#else
static int direct_flag = 0;
#endif

static void print_help(void) {
	printf("Usage: %s [options] device [devices...] [-o report]\n", PACKAGE);
	printf("\n");
	printf("Options\n");
	printf(" -b	Start test at this location.\n");
	printf(" -e	End test after this location.\n");
	printf(" -c	Use memory buffer of this size.\n");
	printf(" -s	Perform no more than this many seeks.\n");
	printf(" -w	Destructive write test.\n");
#ifdef O_DIRECT
	printf(" -d     Do not bypass disk cache via O_DIRECT.\n");
#endif
	printf(" -o	Save output in the file \"report\".\n");
}

static uint64_t get_number(const char *string)
{
	int sz = strlen(string);
	double result = atof(string);
	char last = string[sz - 1];

	switch (last) {
		case 'k':
		case 'K':
			result *= 1024;
			break;
		case 'm':
		case 'M':
			result *= 1048576;
			break;
		case 'g':
		case 'G':
			result *= 1073741824;
			break;
	}

	return result;
}

static uint32_t get_block_size(const char *fname)
{
	struct stat devinfo;
	uint32_t blksz;
	int fd, is_block = 0;

	/* File blocksize? */
	if (!stat(fname, &devinfo)) {
		blksz = devinfo.st_blksize;
		if (S_ISBLK(devinfo.st_mode))
			is_block = 1;
	}

	/* block device, use ioctl to get sector size */
	if (is_block) {
#ifdef BLKSSZGET
		fd = open(fname, O_RDONLY);
		if (fd < 0) {
			perror(fname);
			goto try_again;
		}
		ioctl(fd, BLKSSZGET, &blksz);

		close(fd);
#endif
	}

	if (blksz)
		goto found_blksz;

try_again:
	perror(fname);
	fprintf(stderr, "%s: Unknown block size; guessing page size "
		"alignment?\n", fname);

#ifdef _SC_PAGESIZE
	blksz = sysconf(_SC_PAGESIZE);
	if (blksz > 0)
		goto found_blksz;
#endif
#ifdef PAGESIZE
	blksz = sysconf(PAGESIZE);
	if (blksz > 0)
		goto found_blksz;
#endif
	blksz = 4096;

found_blksz:
	return blksz;
}

static uint64_t get_size(const char *fname)
{
	uint64_t res = 0;
	int is_block = 0;
	int fd;
#ifdef __USE_LARGEFILE64
	struct stat64 si64;
#else
	struct stat si;
#endif

#ifdef __USE_LARGEFILE64
	if (!stat64(fname, &si64)) {
		res = si64.st_size;
		if (S_ISBLK(si64.st_mode))
			is_block = 1;
	}
#else
	if (stat(fname, &si)) {
		res = si.st_size;
		if (S_ISBLK(si.st_mode))
			is_block = 1;
	}
#endif

	if (is_block) {
#ifdef BLKGETSIZE64
		fd = open(fname, O_RDONLY);
		if (fd < 0) {
			perror(fname);
			return 0;
		}
		if (ioctl(fd, BLKGETSIZE64, &res)) {
			perror(fname);
			return 0;
		}
		close(fd);
#else
		if (!res)
			fprintf("Don't know how to get size of block "
				"device!\n");
#endif
	}
	
	return res;
}

static int seed_random(void) {
	int fp;
	long seed;

	fp = open("/dev/urandom", O_RDONLY);
	if (fp < 0) {
		perror("/dev/urandom");
		return 0;
	}

	if (read(fp, &seed, sizeof(seed)) != sizeof(seed)) {
		perror("read random seed");
		return 0;
	}

	close(fp);
	srand(seed);

	return 1;
}

static uint64_t get_randnum_align(uint64_t min, uint64_t max, uint64_t align) {
	return (min + (uint64_t)((double)(max - min) * (rand() / (RAND_MAX + 1.0)))) &
		~(align - 1);
}

static void time_device(const char *dev, uint64_t start, uint64_t stop,
			uint64_t bufsize, FILE *report, uint64_t max_seeks)
{
	FILE *out = stdout;
	uint64_t max_size, pos, newpos, i, oldpos;
	void *buf;
	struct timespec now, before, rep;
	double nowf, beforef, repf;
	int clock_type = CLOCK_MONOTONIC;
	int fd, ret, blksz;

	/* Flip backwards params */
	if (start && stop && (start > stop)) {
		max_size = start;
		start = stop;
		stop = max_size;
	}

	/* Check parameters */
	max_size = get_size(dev);
	if (!max_size) {
		fprintf(stderr, "%s: Size is zero, skipping.\n", dev);
		return;
	}

	if (start > max_size) {
		fprintf(stderr, "%s: Starting at %"PRIu64", which is beyond the "
			"end of the device at %"PRIu64"!\n", dev, start, max_size);
		return;
	}

	fprintf(out, "%s: Max size %"PRIu64"\n", dev, max_size);

	if (stop > max_size) {
		fprintf(stderr, "%s: Clamping end to %"PRIu64".\n", dev, max_size);
		stop = max_size;
	} else if (!stop) {
		stop = max_size;
	}

	blksz = get_block_size(dev);
	fprintf(out, "%s: Block size %d bytes.\n", dev, blksz);

	if (!bufsize)
		bufsize = blksz;

	if (!max_seeks)
		max_seeks = (stop - start) / (blksz * 1000);
	
	if (report == stdout)
		out = tty;
	else
		fprintf(out, "%s: Seeking %"PRIu64" times from %"PRIu64" to %"PRIu64" in "
			"%"PRIu64" byte chunks.\n", dev, max_seeks, start, stop,
			bufsize);
	fprintf(report, "%s: Seeking %"PRIu64" times from %"PRIu64" to %"PRIu64" in %"PRIu64" byte "
		"chunks.\n", dev, max_seeks, start, stop, bufsize);

	/* grab buffer */
	if (posix_memalign(&buf, blksz, bufsize)) {
		perror("posix_memalign");
		return;
	}

	/* lock pages */
#ifdef _POSIX_MEMLOCK_RANGE
	if (mlock(buf, bufsize)) {
		perror("mlock");
		fprintf(stderr, "Can't lock pages; proceeding anyway.\n");
	}
#else
#warn mlock not supported
#endif

	/* open device */
	fd = open(dev, access_flag | direct_flag | O_LARGEFILE);
	if (!fd) {
		perror("dev");
		goto unlock_mem;
	}

	/* make sure CLOCK_MONOTONIC is supported */
	if (clock_gettime(clock_type, &now)) {
		clock_type = CLOCK_REALTIME;
	}

	/* seek to starting point */
	pos = lseek64(fd, (newpos = get_randnum_align(start, stop, bufsize)),
		      SEEK_SET);
	if (pos != newpos) {
		perror("lseek");
		goto end_test;
	}

	clock_gettime(clock_type, &rep);
	repf = (rep.tv_sec) + ((double)rep.tv_nsec / 1000000000);

	/* loop */
	for (i = max_seeks; i > 0; i--) {
		clock_gettime(clock_type, &before);
		oldpos = newpos;
		newpos = get_randnum_align(start, stop, bufsize);
		pos = lseek64(fd, newpos, SEEK_SET);
		if (pos != newpos) {
			perror("lseek");
			break;
		}
		ret = io_func(fd, buf, bufsize);
		if (!ret)
			break;
		else if (ret < 0) {
			perror(dev);
			if (ret != EIO)
				break;
		}
		clock_gettime(clock_type, &now);

		nowf = (now.tv_sec) + ((double)now.tv_nsec / 1000000000);
		beforef = (before.tv_sec) +
			  ((double)before.tv_nsec / 1000000000);
		if (newpos > oldpos)
			pos = newpos - oldpos;
		else
			pos = oldpos - newpos;
		fprintf(report, "%"PRIu64", %"PRIu64", %"PRIu64", %.2f\n", oldpos, newpos,
			pos,
			(nowf - beforef) * 1000);
		if ((out == stdout) && ((nowf - repf) > 5)) {
			fprintf(out, "%s: %"PRIu64" seeks (%.2f%%)            \r",
				dev, max_seeks - i,
				((double)1.0 - ((double)i / max_seeks)) * 100);
			fflush(out);
			repf = nowf;
		}
	}

	if (out == stdout)
		fprintf(out, "%s: %"PRIu64" seeks (%.2f%%)            \n",
			dev, max_seeks - i,
			((double)1.0 - ((double)i / max_seeks)) * 100);

end_test:
	/* close device */
	close(fd);

unlock_mem:
	/* unlock pages */
#ifdef _POSIX_MEMLOCK_RANGE
	if (munlock(buf, bufsize)) {
		perror("munlock");
	}
#endif
	free(buf);
}

/* Slurp files from linux/ioprio.h */
#ifdef SYS_ioprio_set
 #ifndef IOPRIO_PRIO_VALUE
enum {
	IOPRIO_CLASS_NONE,
	IOPRIO_CLASS_RT,
	IOPRIO_CLASS_BE,
	IOPRIO_CLASS_IDLE,
};

enum {
        IOPRIO_WHO_PROCESS = 1,
        IOPRIO_WHO_PGRP,
        IOPRIO_WHO_USER,
};

#define IOPRIO_CLASS_SHIFT		(13)
#define IOPRIO_PRIO_VALUE(class, data)	(((class) << IOPRIO_CLASS_SHIFT) | data)

static long ioprio_set(int which, int who, int ioprio)
{
	return syscall(SYS_ioprio_set, which, who, ioprio);
}
 #endif
#endif

static void bump_priority(void)
{
	struct sched_param sp;
	int res;

	res = ioprio_set(IOPRIO_WHO_PROCESS, getpid(),
			 IOPRIO_PRIO_VALUE(IOPRIO_CLASS_RT, 7));
	if (res)
		perror("ioprio_set");

	sp.sched_priority = sched_get_priority_max(SCHED_FIFO);
	res = sched_setscheduler(0, SCHED_FIFO, &sp);
	if (res)
		perror("sched_setscheduler");
}

int main(int argc, char *argv[])
{
	uint64_t start, stop, bufsize, max_seeks;
	int i;
	FILE *report = stdout;
	char c;

	start = stop = 0;
	max_seeks = bufsize = 0;

	tty = fopen("/dev/tty", "a+");
	if (!tty) {
		perror("/dev/tty");
		return 1;
	}
	fprintf(tty, "%s %s, Copyright (C) 2006-2009 Darrick Wong.\n",
		PACKAGE, PACKAGE_VERSION);

	/* parse args */
	while((c = getopt(argc, argv, "wb:e:c:o:s:d")) != -1) {
		switch(c) {
			case 'o':
				report = fopen(optarg, "w+");
				if (!report) {
					perror(optarg);
					return 2;
				}
				break;
			case 'b':
				start = get_number(optarg);
				break;
			case 'e':
				stop = get_number(optarg);
				break;
			case 'c':
				bufsize = get_number(optarg);
				break;
			case 's':
				max_seeks = get_number(optarg);
				break;
			case 'd':
				direct_flag = 0;
				break;
			case 'w':
				io_func = (io_func_t)write;
				access_flag = O_WRONLY;
				break;
			default:
				print_help();
				return 1;
		}
	}

	if (argc == optind) {
		print_help();
		return 1;
	}

	if (!seed_random())
		return 2;

	bump_priority();

	for (i = optind; i < argc; i++) {
		time_device(argv[i], start, stop, bufsize, report, max_seeks);
	}

	return 0;
}
